"""
Plugins allowing to check licenses and compare with Gamma's list of allowed licenses.

to run this plugin, let's say BSD are unknown for this exemple:

.. code-block:: bash

   gamma check-licenses
     ____
    / ___| __ _ _ __ ___  _ __ ___   __ _
    | |  _ / _` | '_ ` _ \| '_ ` _ \ / _` |
    | |_| | (_| | | | | | | | | | | | (_| |
    \____|\__,_|_| |_| |_|_| |_| |_|\__,_|


    gamma-cli (0.3.0) is currently in beta
    please report any bug on: https://github.gamma.bcg.com/BCG/gamma-cli/issues

    analyze done using python located at: /Users/freierniels/work/gamma/tools/gamma-cli/.env/bin/python, if you want to use python from your virtual env, please add the option '--python', ex: '--python `which python`'
    Dependencies with problem
    ├── ! commitizen MIT License - Python commitizen client tool
    │   └── ? packaging Apache Software License, BSD License - Core utilities for Python packages

"""

import sys
import shutil

import click
from rich import print

from gamma.cli.services import LicenceService


@click.command()
@click.pass_obj
@click.option(
    "-a",
    "--all",
    "all",
    show_default=True,
    default=False,
    is_flag=True,
    help="Will display all dependencies, including the allowed",
)
@click.option(
    "-l",
    "--no-legend",
    "legend",
    show_default=True,
    default=True,
    is_flag=True,
    help="Will display the legend",
)
@click.option(
    "-w",
    "--fail-on-unknown",
    "fail_unknown",
    show_default=True,
    default=False,
    is_flag=True,
    help="Will exit with an error code in case of unknown license",
)
@click.option(
    "--licenses_settings",
    "licenses_settings",
    help="path of a custom licenses settings file, used for offline scenario",
)
@click.option(
    "-p",
    "--python",
    "python_path",
    help="Path of the python you want to use, useful to handle virtual env",
)
@click.option(
    "-j",
    "--json",
    "json_path",
    help="Path of the json output file",
)
def cli(
    telemetry, all, legend, python_path, fail_unknown, licenses_settings, json_path
):
    """ Checks licenses of the package in the current env """

    if telemetry is not None:
        telemetry.track_event("check-licences")

    if python_path:
        cmd_exists = lambda x: shutil.which(x) is not None
        if not cmd_exists(python_path):
            print(f"[red]the path to python is not valid: {python_path}")
            sys.exit(1)
    service = LicenceService(
        python_path or shutil.which("python") or shutil.which("python3"),
        True if sys.executable != python_path else False,
        licenses_settings,
        telemetry,
    )
    return_code = service.analyse(fail_unknown)
    service.print(all, legend, json_path)
    sys.exit(return_code)
