"""
Licenses services used to list and confirm if the packages can be used for Gamma projects
"""

from dataclasses import dataclass
import inspect
import subprocess
import sys
from pathlib import Path
from typing import Dict, List, Optional, ForwardRef
import json
import warnings
from operator import attrgetter
from enum import Enum
from itertools import chain
import requests

flatten = chain.from_iterable
from collections.abc import Mapping
from collections import OrderedDict

from dataclasses_json import dataclass_json, Undefined
from marshmallow import Schema, fields
import yaml
from rich import print
from rich.text import Text
from rich.tree import Tree
from rich.table import Table
from rich.console import Console
from git.repo.base import Repo
from applicationinsights import TelemetryClient


from gamma.cli.services.github import GitHubService

LICENSES_VERSION_URL = "https://gammatools.blob.core.windows.net/licenses/latest"


class PackageValidity(Enum):
    ALLOWED = 1
    DENIED = 2
    IMPACTED_DENIED = 3
    IMPACTED_UNKNOWN = 4
    UNKNOWN = 5
    EXCEPTION = 6


symbols = {
    PackageValidity.ALLOWED: ["✔", "green", "allowed"],
    PackageValidity.EXCEPTION: ["!", "yellow", "exception"],
    PackageValidity.IMPACTED_DENIED: ["!", "red", "denied by impact"],
    PackageValidity.DENIED: ["✘", "red", "denied"],
    PackageValidity.IMPACTED_UNKNOWN: ["!", "blue", "unknown by impact"],
    PackageValidity.UNKNOWN: ["?", "blue", "unknown"],
}


def sorted_tree(tree):
    """Sorts the dict representation of the tree
    The root packages as well as the intermediate packages are sorted
    in the alphabetical order of the package names.
    :param dict tree: the pkg dependency tree obtained by calling
                     `construct_tree` function
    :returns: sorted tree
    :rtype: collections.OrderedDict
    """
    step = sorted(
        [(k, sorted(v, key=attrgetter("key"))) for k, v in tree.items()],
        key=lambda kv: kv[0].key,
    )
    return OrderedDict(step)


class PackageDAG(Mapping):
    @classmethod
    def from_pkgs(cls, pkgs):
        idx = {p.key: p for p in pkgs}
        m = {}
        for p in pkgs:
            m[p] = []
            for dep in p.dependencies:
                if dep in idx:
                    m[p].append(idx[dep])

        d = cls(m)
        for k, value in d.items():
            k.dependencies = value

        return d

    def __init__(self, m):
        self._obj = m
        self._index = {p.key: p for p in list(self._obj)}

    def sort(self):
        """Return sorted tree in which the underlying _obj dict is an
        OrderedDict, sorted alphabetically by the keys
        :returns: Instance of same class with OrderedDict
        """
        return self.__class__(sorted_tree(self._obj))

    # Methods required by the abstract base class Mapping
    def __getitem__(self, *args):
        return self._obj.get(*args)

    def __iter__(self):
        return self._obj.__iter__()

    def __len__(self):
        return len(self._obj)


class LicenseConfiguration:

    allowed: Dict[str, bool]
    denied: Dict[str, bool]
    exceptions: Dict[str, bool]
    unknown: str

    def __init__(self, **entries):
        try:
            self.allowed = {i.lower(): True for i in entries["allowed"]}
            self.denied = {i.lower(): True for i in entries["denied"]}
            self.exceptions = {i.lower(): True for i in entries["exceptions"]}
            self.unknown = entries["unknown"]
        except KeyError as ex:
            print(f"[red]invalid format for tge licenses setting file")
            sys.exit(1)


@dataclass_json
@dataclass
class Package(Schema):
    key: str
    package_name: str
    installed_version: str
    dependencies: Optional[List[ForwardRef("Package")]] = None  # type: ignore
    required_version: Optional[str] = None
    license: Optional[str] = None
    description: Optional[str] = None
    status: Optional[PackageValidity] = PackageValidity.UNKNOWN

    def __hash__(self):
        return hash((self.key, self.license))


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class CondaPackage(Schema):
    name: str
    version: str
    depends: List[str]
    license: str = "UNKNOWN"


class LicenceService:
    _config: LicenseConfiguration
    _python_path: str
    _custom_python: bool
    _packages_with_license: List[Package]
    _packages_with_license_as_dict: Dict[str, Package]
    _packages_with_deps: List[Package]
    _conda_packages_with_deps: Optional[List[Package]] = None
    _telemetry: Optional[TelemetryClient]

    def __init__(
        self,
        python_path: str,
        custom_python: bool,
        licenses_settings: Optional[str] = None,
        telemetry: Optional[TelemetryClient] = None,
    ):
        self._python_path = python_path
        self._custom_python = custom_python
        self._telemetry = telemetry
        self._init(licenses_settings)

    def _init(self, licenses_settings_path: Optional[str] = None):
        try:
            if licenses_settings_path:
                with open(str(licenses_settings_path), "r") as f:
                    self._config = LicenseConfiguration(**yaml.safe_load(f))
            else:
                self._update_gamma_licenses()
                self._config = self._default_license_configuration()
        except Exception as ex:
            print(f"[red]{ex}")
            sys.exit(1)

    def _update_gamma_licenses(self) -> None:
        """Check and update  gamma-licenses package"""
        console = Console()

        with console.status("gamma-licenses: Checking for updates..."):
            r = requests.get(LICENSES_VERSION_URL)
            r.raise_for_status()
            _, remote_version, url = r.text.strip().split()
            try:
                from gamma.licenses import __version__

                # semver parsing
                this = tuple(map(int, __version__.split("-", 1)[0].split(".")))
                that = tuple(map(int, remote_version.split("-", 1)[0].split(".")))
                upgrade = this < that
            except ModuleNotFoundError:
                upgrade = True

        if upgrade:
            msg = f"[cyan]Upgrading gamma-licenses to [yellow]{remote_version}..."
            with console.status(msg):
                cmd = [sys.executable, "-m", "pip", "install", "-q", "-U", url]
                subprocess.run(
                    cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True
                )

    def _default_license_configuration(self) -> LicenseConfiguration:
        """Return default license configuration from gamma-licenses"""
        from gamma.licenses import get_licenses_yaml
        from io import StringIO

        content = get_licenses_yaml()
        stream = StringIO(content)
        stream.seek(0)
        return LicenseConfiguration(**yaml.safe_load(stream))

    def analyse(self, fail_unknown: bool) -> int:
        console = Console()

        with console.status("[bold green]Computing the licenses from conda..."):
            conda_path = Path(self._python_path).parent.parent / "conda-meta"
            if conda_path.exists() and conda_path.is_dir():
                conda_packages = self._extract_conda_packages(conda_path)
                conda_packages = self._flag_licenses(conda_packages)
                conda_tree = self._analyse_conda_packages(conda_packages)
                self._packages_with_license_as_dict = {i.key: i for i in conda_tree}
                self._conda_packages_with_deps = self._explore(conda_tree)

        with console.status("[bold green]Computing the licenses from pip..."):
            self._packages_with_license = self._run_pip_licenses()
            self._packages_with_license = self._flag_licenses(
                self._packages_with_license
            )
            self._packages_with_license_as_dict = {
                i.key: i for i in self._packages_with_license
            }
            self._packages_with_deps = self._run_pipdeptree()
            self._packages_with_deps = self._explore(self._packages_with_deps)

        return_code = 0
        for package in self._packages_with_deps:
            if package.status not in [
                PackageValidity.ALLOWED,
                PackageValidity.UNKNOWN,
                PackageValidity.IMPACTED_UNKNOWN,
            ]:
                return_code = 1
                break
            elif fail_unknown and package.status == PackageValidity.UNKNOWN:
                return_code = 1
                break

        return return_code

    def _filter_conda_system(self, filename) -> bool:
        if filename.startswith(("_", "ld_impl_linux", "python-")):
            return False
        return True

    def _extract_conda_packages(self, conda_meta_path: Path) -> List[Package]:
        ret: List[Package] = []

        jsons = Path(conda_meta_path).glob("*.json")
        packages = []
        for json_metadata in jsons:
            if self._filter_conda_system(json_metadata.name):
                with open(str(json_metadata), "r") as f:
                    conda_package = CondaPackage.schema().loads(f.read())  # type: ignore
                    deps = []
                    for dep in conda_package.depends:
                        deps.append(dep.split(" ")[0])
                    packages.append(
                        Package(
                            key=conda_package.name.lower(),
                            license=conda_package.license,
                            package_name=conda_package.name,
                            installed_version=conda_package.version,
                            dependencies=deps,
                        )
                    )
        return packages

    def _analyse_conda_packages(self, packages) -> List[Package]:
        dag = PackageDAG.from_pkgs(packages)
        tree = dag.sort()
        branch_keys = set(r.key for r in flatten(tree.values()))
        nodes = [p for p in tree.keys() if p.key not in branch_keys]
        return nodes

    def _walk_print(self, tree: Tree, dict_tree, package: Package, all: bool):
        if all or package.status != PackageValidity.ALLOWED:
            if not package.description:
                tree = tree.add(
                    f"[{symbols[package.status][1]}]{symbols[package.status][0]}[/{symbols[package.status][1]}] {package.package_name} [dim]{package.license}[/dim]"  # type: ignore
                )
            else:
                tree = tree.add(
                    f"[{symbols[package.status][1]}]{symbols[package.status][0]}[/{symbols[package.status][1]}] {package.package_name} [dim]{package.license} - {package.description}[/dim]"  # type: ignore
                )
            dict_package = {
                "name": package.package_name,
                "license": package.license,
                "description": package.description,
                "status": symbols[package.status][2],  # type: ignore
            }

            if self._telemetry:
                self._telemetry.track_event(
                    "python_package",
                    {
                        "name": package.package_name,
                        "license": package.license,
                        "status": symbols[package.status][2],  # type: ignore
                    },
                )

            if package.dependencies:
                dict_package["deps"] = []  # type: ignore
                for dep in package.dependencies:
                    self._walk_print(tree, dict_package["deps"], dep, all)
            dict_tree.append(dict_package)

    def print(self, all: bool, legend: bool, json_path: Optional[str]):
        if self._custom_python:
            print(f"analyze done using python located at: {self._python_path}")
        else:
            print(
                f"analyze done using python located at: {self._python_path}, if you want to use python from your virtual env, please add the option '--python', ex: '--python `which python`'"
            )
        label = (
            "dependencies requiring attention" if not all else "complete list of dependencies"
        )

        dict_conda_tree = []  # type: ignore
        if self._conda_packages_with_deps:
            tree = Tree(f"Conda {label}")
            for package in self._conda_packages_with_deps:
                if all or package.status != PackageValidity.ALLOWED:
                    self._walk_print(tree, dict_conda_tree, package, all)
            print("")
            print(tree)

        tree = Tree(f"Pip {label}")
        dict_pip_tree = []  # type: ignore
        for package in self._packages_with_deps:
            if all or package.status != PackageValidity.ALLOWED:
                self._walk_print(tree, dict_pip_tree, package, all)
        print("")
        print(tree)

        if json_path:
            output = {"pip": dict_pip_tree}
            if len(dict_conda_tree):
                output["conda"] = dict_conda_tree
            with open(json_path, "w") as f:
                f.write(json.dumps(output, indent=2))

        if legend:
            legend_table = Table(title="Legend")
            legend_table.add_column("Symbol", justify="center")
            legend_table.add_column("Description", justify="left")
            legend_table.add_column("Action", justify="left")
            legend_table.add_row(
                f"[{symbols[PackageValidity.ALLOWED][1]}]{symbols[PackageValidity.ALLOWED][0]}[/{symbols[PackageValidity.ALLOWED][1]}]",
                "Approved",
                "Everything is OK.",
            )
            legend_table.add_row(
                f"[{symbols[PackageValidity.DENIED][1]}]{symbols[PackageValidity.DENIED][0]}[/{symbols[PackageValidity.DENIED][1]}]",
                "Avoid",
                "only use this package if you have exhausted all options to substitute it - requires you to clarify legal/IP implications with DEVEX and GAMMA legal",
            )
            legend_table.add_row(
                f"[{symbols[PackageValidity.EXCEPTION][1]}]{symbols[PackageValidity.EXCEPTION][0]}[/{symbols[PackageValidity.EXCEPTION][1]}]",
                "Exception",
                "OK if not distributed and used for dev only, ex: pylint, black, ...",
            )
            legend_table.add_row(
                f"[{symbols[PackageValidity.UNKNOWN][1]}]{symbols[PackageValidity.UNKNOWN][0]}[/{symbols[PackageValidity.UNKNOWN][1]}]",
                "Unknown",
                "Unknown license type, please manually check license. Requires review if not explicitly allowed.",
            )
            legend_table.add_row(
                f"[{symbols[PackageValidity.IMPACTED_DENIED][1]}]{symbols[PackageValidity.IMPACTED_DENIED][0]}[/{symbols[PackageValidity.IMPACTED_DENIED][1]}]",
                "Avoid due to sub-dependency",
                "Sub-dependency has a copyleft license, only use this package if you have exhausted all options to substitute it - requires you to clarify legal/IP implications with DEVEX and GAMMA legal",
            )
            legend_table.add_row(
                f"[{symbols[PackageValidity.IMPACTED_UNKNOWN][1]}]{symbols[PackageValidity.IMPACTED_UNKNOWN][0]}[/{symbols[PackageValidity.IMPACTED_UNKNOWN][1]}]",
                "Unknown due to sub-dependency",
                "Sub-dependency requires manual check.",
            )
            print(legend_table)

        print("")
        print("[yellow]!! check-licenses is currently in beta !!")
        print("")
        print("[bold magenta]For license review, please contact the case [bold yellow]Devex coach[bold magenta], or reach out to [bold yellow]devex@bcg.com[bold magenta]")

        if self._telemetry:
            self._telemetry.flush()

    def _flag_licenses(self, packages: List[Package]) -> List[Package]:
        for package in packages:
            if package.key in self._config.exceptions:
                package.status = PackageValidity.EXCEPTION
            elif package.license.lower() in self._config.denied:  # type: ignore
                package.status = PackageValidity.DENIED
            elif package.license.lower() in self._config.allowed:  # type: ignore
                package.status = PackageValidity.ALLOWED
            else:
                package.status = PackageValidity.UNKNOWN
        return packages

    def _assign_info(self, package: Package):
        if package.key in self._packages_with_license_as_dict:
            package.license = self._packages_with_license_as_dict[package.key].license
            package.description = self._packages_with_license_as_dict[
                package.key
            ].description
            package.status = self._packages_with_license_as_dict[package.key].status

    def _walk(self, package: Package, level: int = 0) -> PackageValidity:
        self._assign_info(package)
        # print(level*"\t", package.package_name)
        if package.dependencies:
            for dep in package.dependencies:
                self._assign_info(dep)
                dep_validity = self._walk(dep, level + 1)
                if dep_validity in [
                    PackageValidity.UNKNOWN,
                    PackageValidity.IMPACTED_UNKNOWN,
                ] and package.status not in [
                    PackageValidity.EXCEPTION,
                    PackageValidity.IMPACTED_DENIED,
                ]:
                    package.status = PackageValidity.IMPACTED_UNKNOWN
                elif (
                    dep_validity
                    in [PackageValidity.DENIED, PackageValidity.IMPACTED_DENIED]
                    and package.status != PackageValidity.EXCEPTION
                ):
                    package.status = PackageValidity.IMPACTED_DENIED

        return package.status  # type: ignore

    def _explore(self, packages: List[Package]) -> List[Package]:
        for package in packages:
            self._walk(package)
        return packages

    def _run_command(self, file_path: str, args: str) -> str:
        cmd = [self._python_path, file_path]
        cmd.extend(args.split(" "))
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        stdout, stderr = out.communicate()
        return stdout.decode("utf-8")

    def _run_pip_licenses(self) -> List[Package]:
        packages = []
        file_path = (
            Path(inspect.getsourcefile(sys.modules[__name__])).parent  # type: ignore
            / "helpers"
            / "piplicenses.py"
        )
        args = "--from=mixed --with-system --format=json --with-description"
        ret = self._run_command(str(file_path), args)
        for line in ret.splitlines():
            data = line.split("|||")
            packages.append(
                Package(
                    key=data[0].lower(),
                    package_name=data[0],
                    license=data[2],
                    installed_version=data[1],
                    description=data[3],
                )
            )
        return packages

    def _run_pipdeptree(self):
        file_path = (
            Path(inspect.getsourcefile(sys.modules[__name__])).parent
            / "helpers"
            / "pipdeptree.py"
        )
        args = "--json-tree"
        ret = self._run_command(file_path, args)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            packages = Package.schema().loads(ret, many=True)
            return packages
