import pathlib as pl
import platform
import sys
import logging
from typing import List

import setuptools
from setuptools.command.install import install

LOGGER = logging.getLogger(__name__)


def _sublines_in_lines(sublines: List[str], lines: List[str]) -> bool:
    """Naive function to find subarray in an array of strings"""
    i = 0
    j = 0
    while i < len(lines) and j < len(sublines):
        if sublines[j] == lines[i]:
            i = i + 1
            j = j + 1
            if j == len(sublines) - 1:
                return True
        else:
            i = i - j + 1
            j = 0
    return False


def _correct_non_tty():
    """Correction for Git Bash

    Git bash is piping a bash terminal to the user so it runs in non-TTY mode.
    To avoid related problems (CLI freeze at inputs, color rendering) we need to run gamma usin 'winpty'.

    This function exports the line "alias gamma='winpty gamma'" to the user's .bash_profile if he's installing in a non-TTY teminal
    """

    if platform.system() == "Windows" and not (
        sys.stdin.isatty() and sys.stdout.isatty()
    ):
        config_path = pl.Path(pl.Path.home(), ".bash_profile")
        config_lines = [
            "\n",
            "# >>> gamma initialize\n",
            "# !! Contents within this block are managed by gamma-cli and should not be modified\n",
            "alias gamma='winpty gamma'\n",
            "# <<< gamma initialize\n",
        ]
        # We check if we need to write the config
        # The only case in wich we don't need to write it is when '$HOME/.bash_profile' exists and already contains the config
        write_config = True
        if config_path.is_file():
            with open(config_path, "r") as fr:
                bash_profile = fr.readlines()
            if _sublines_in_lines(config_lines, bash_profile):
                write_config = False
        # config is written if required
        if write_config:
            logging.debug(
                "You are running a Windows non-TTY shell. We need to add config to your .bash_profile"
            )
            with open(config_path, "a") as fw:
                fw.writelines(config_lines)
            LOGGER.warning(
                "We added a configuration for gamma-cli in your .bash_profile"
            )
            LOGGER.warning(
                "You need to restart your shell or to run `source ~/.bash_profile` before using gamma-cli"
            )


class PostInstallCommand(install):
    """This class override setuptools install to run post install scripts"""

    def run(self):
        install.run(self)
        _correct_non_tty()


with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="gamma-cli",
    version="0.4.1",
    entry_points={"console_scripts": ["gamma=gamma.cli.gamma:cli"]},
    author="Niels Freier",
    author_email="freier.niels@bcg.com",
    description="Gamma toolbox",
    long_description=long_description,
    include_package_data=True,
    python_requires=">=3.7",
    long_description_content_type="text/markdown",
    url="https://github.gamma.bcg.com/BCG/gamma-cli",
    install_requires=requirements,
    license="Apache License 2.0",
    packages=setuptools.find_namespace_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    cmdclass={"install": PostInstallCommand},
)
