import tempfile
from pathlib import Path
import shutil
from subprocess import run
import pytest


def docker():
    if shutil.which("docker"):
        return "docker"
    elif shutil.which("podman"):
        return "podman"
    else:
        raise Exception("Missing docker")


# When using windows containers, use same network as WSL
WIN_DOCKER_NETWORK = "WSL"


@pytest.mark.install
def test_setup_ubuntu():
    """Test install script.
    Requires docker. Happy path in Ubuntu container"""

    image = "ubuntu:20.04"
    commands = """
    export DEBIAN_FRONTEND=noninteractive
    apt update
    apt install -y python3.8 python3-pip python3.8-venv git
    python3 /app/install
    export PATH="$HOME/.local/bin:${PATH}"
    pipx --version
    gamma --version
    """

    with tempfile.TemporaryDirectory() as td:
        ptd = Path(td)
        script = ptd / "install"
        cmd = ptd / "commands"
        script_content = Path("scripts/install.py").read_text()
        script.write_text(script_content)
        cmd.write_text(commands)
        docker_cmd = [docker(), "run", "--rm", "-i", "-v", f"{td}:/app", image]
        docker_cmd += ["bash", "-e", "-x", "/app/commands"]
        run(docker_cmd, check=True)


@pytest.mark.win_install
def test_setup_windows():
    """Test install script.
    Requires docker windows containers.
    """

    image = "mcr.microsoft.com/windows:2004"
    commands = (Path(__file__).parent / "commands.ps1").read_text()
    script_content = Path("scripts/install.py").read_text()

    with tempfile.TemporaryDirectory() as td:
        ptd = Path(td)
        script = ptd / "install"
        cmd = ptd / "commands.ps1"
        script.write_text(script_content)
        cmd.write_text(commands)
        docker_cmd = [docker(), "run", "--net", WIN_DOCKER_NETWORK]
        docker_cmd += ["--rm", "-i", "-v", f"{td}:c:\\app", image]
        docker_cmd += ["powershell", "-ExecutionPolicy", "ByPass"]
        docker_cmd += ["-file", "c:\\app\\commands.ps1"]
        print(" ".join(docker_cmd))
        run(docker_cmd, check=True)